////////////////////////////////////////////////////////////////////////////////
//
//	calib.c
//
//	Alain BARTHE
//
//	Shared library to access solarOrbiter calibrations from Python
//
////////////////////////////////////////////////////////////////////////////////

#define	PY_SSIZE_T_CLEAN

#define NPY_NO_DEPRECATED_API NPY_1_7_API_VERSION

#include <stdlib.h>
#include <stddef.h>

#include <Python.h>
#include "numpy/arrayobject.h"
#include "datetime.h"

#include "TOOLS.h"
#include "CALIB.h"

#define	LOGFILE		"calib.log"
#define	MAX_DIMS	6

static PyArray_Descr *	energy_bin_descr;	// energy bin descriptor
static PyArray_Descr *	angle_bin_descr;	// angle bin descriptor
static PyArray_Descr *	float_descr;		// float descriptor

static PyTypeObject *	recarray_type;		// numpy.recarray type

typedef struct {				// calibration table structure
	char *		key;			// key described in catalog
	PyObject *	object;			// numpy array object
	void *		data;			// C global variable array
	PyArray_Descr **type;			// numpy dtype (NPY_FLoAT)
	int		ndim;			// number of array dimension
	npy_intp	dims [MAX_DIMS];	// sizes of each dimension
}	t_calib_table;

static	t_calib_table	tables [] = {
	{ "AZS",	NULL,	& AZS,	& angle_bin_descr,	1, { 11 }},
	{ "AZF",	NULL,	& AZF, 	& angle_bin_descr,	2, { 11, 9 }},
	{ "ELS",	NULL, 	& ELS,	& angle_bin_descr,	1, { 9 }},
	{ "ELF",	NULL, 	& ELF,	& angle_bin_descr,	2, { 11, 9 }},
	{ "ELD",	NULL,	& ELD,	& float_descr,		1, { 96 }},
	{ "EVS",	NULL,	& EVS,	& energy_bin_descr,	1, { 96 }},
	{ "EVF",	NULL,	& EVF,	& energy_bin_descr,	3, { 11, 9, 96 }},
	{ "GFF",	NULL,	& GFF,	& float_descr,		2, { 11, 9 }},
	{ "CNF",	NULL, 	& CNF,	& float_descr, 		3, { 11, 9, 96 }},
	{ "YIL",	NULL,	& YIL,	& float_descr,		4, { 11, 9, 11, 9}},
	{ NULL,		NULL,	NULL,	NULL,			0, { 0 }}
};


////////////////////////////////////////////////////////////////////////////////
//
//	Search calibration key entry
//
static 	
t_calib_table * search_key (char * key)
{
	t_calib_table *	ptr;
	
	for (ptr = tables; ptr->key != NULL; ptr++)
		if (strcmp (key, ptr->key) == 0) return ptr;
	return NULL;
}


////////////////////////////////////////////////////////////////////////////////
//
//	Create float descriptor
//
static 
PyArray_Descr * make_float_descr (void)
{
	return PyArray_DescrFromType (NPY_FLOAT);
}


////////////////////////////////////////////////////////////////////////////////
//
//	Create energy bin descriptor
static 
PyArray_Descr * make_energy_bin_descr (void)
{
	PyArray_Descr *	descr;
	PyObject *	dict;
	
	dict = Py_BuildValue ("{s: [s, s, s], s: [s, s, s], s: [i, i, i], s: i}",
		"names",
			"emin",
			"emoy",
			"emax",
		"formats",
			"f4",
			"f4",
			"f4",
		"offsets",
			offsetof (t_energy_bin, min),
			offsetof (t_energy_bin, moy),
			offsetof (t_energy_bin, max),
		"itemsize",
			 sizeof (t_energy_bin)
	);

	if (dict == NULL) {
		PyErr_SetString (PyExc_TypeError, "Cannot build energy_bin dictionnary");
		return NULL;
	}

	if (PyArray_DescrConverter (dict, & descr) == NPY_FAIL) {
		PyErr_SetString (PyExc_TypeError, "Failed to build energy bin descriptor");
		Py_DECREF (dict);
		return NULL;
	}

	Py_DECREF (dict);

	return descr;
}


////////////////////////////////////////////////////////////////////////////////
//
//	Create angle bin descriptor
static 
PyArray_Descr * make_angle_bin_descr (void)
{
	PyArray_Descr *	descr;
	PyObject *	dict;
	
	dict = Py_BuildValue ("{s: [s, s, s, s, s], s: [s, s, s, s, s], s: [i, i, i, i, i], s: i}",
		"names",
			"amin",
			"amoy",
			"amax",
			"sin",
			"cos",
		"formats",
			"f4",
			"f4",
			"f4",
			"f4",
			"f4",
		"offsets",
			offsetof (t_angle_bin, min),
			offsetof (t_angle_bin, moy),
			offsetof (t_angle_bin, max),
			offsetof (t_angle_bin, sin),
			offsetof (t_angle_bin, cos),
		"itemsize",
			sizeof (t_angle_bin)
	);

	if (dict == NULL) {
		PyErr_SetString (PyExc_TypeError, "Cannot build angle_bin dictionnary");
		return NULL;
	}

	if (PyArray_DescrConverter (dict, & descr) == NPY_FAIL) {
		PyErr_SetString (PyExc_TypeError, "Failed to build angle_bin descriptor");
		Py_DECREF (dict);
		return NULL;
	}

	Py_DECREF (dict);

	return descr;
}


////////////////////////////////////////////////////////////////////////////////
//
//	Return nump.recarray array type
//
static
PyTypeObject * get_recarray_type (void)
{
	PyTypeObject *	tp = NULL;
	PyObject *	np = NULL;

	if ((np = PyImport_ImportModule ("numpy")) == NULL) {

		PyErr_SetString (PyExc_TypeError, "Cannot import 'numpy' module");
		return NULL;;
	}

	if ((tp = (PyTypeObject *) PyObject_GetAttrString (np, "recarray")) == NULL) {

		PyErr_SetString (PyExc_TypeError, "Cannot get 'recarrray' attr");
		Py_DECREF(np);
		return NULL;
	}

	Py_DECREF (np);

	return tp;
}


////////////////////////////////////////////////////////////////////////////////
//
//	Init calibration catalog
//
static 
PyObject * calib_init (PyObject * self, PyObject * args)
{	
	t_err		error = OK;
	t_calib_table *	ptr;

	if (! PyArg_ParseTuple (args, ""))
		return NULL;

	for (ptr = tables; ptr->key != NULL; ptr++) {
		ptr->object = NULL;
	}

	if (Error (error = Set_logfile (LOGFILE))) goto EXIT;

	if (Error (error = Init_calib ("PFM"))) {
	
		PyErr_SetString (PyExc_RuntimeError, "Error in calibration environment");
		goto EXIT;
	}
	if ((angle_bin_descr = make_angle_bin_descr()) == NULL) goto EXIT;

	if ((energy_bin_descr = make_energy_bin_descr()) == NULL) goto EXIT;

	if ((float_descr = make_float_descr()) == NULL) goto EXIT;

	if ((recarray_type = get_recarray_type()) == NULL) goto EXIT;

	Py_INCREF (Py_None);

	return Py_None;

EXIT:	return NULL;
}


////////////////////////////////////////////////////////////////////////////////
//
//	Convert python datetime to Unix timestamp
//
static
time_t	PyDateTime_to_time (PyObject * obj)
{
	return (time_t) PyFloat_AsDouble (PyObject_CallMethod (obj, "timestamp", ""));
}


////////////////////////////////////////////////////////////////////////////////
//
//	Read calibration data for a given timetag and calibration key
//
static 
PyObject * calib_get (PyObject * self, PyObject * args)
{
	char *		key;
	PyObject *	obj;
	time_t		time;
	t_err		error = OK;
	t_calib_table *	ptr;

	if (! PyArg_ParseTuple (args, "Os", & obj, & key)) 
		return NULL;

	if (! PyDateTime_Check (obj)) {
		PyErr_SetString (PyExc_TypeError, "Argument must be a datetime.datetime");
		goto EXIT;
	}

	time = PyDateTime_to_time (obj);

	//	Read calibration table content and feed C global variable
	//
	if (Error (error = Read_calib (key, time))) {

		PyErr_SetString (PyExc_ValueError, "No calibration for key");
		goto EXIT;
	}

	if ((ptr = search_key (key)) == NULL) {
		PyErr_SetString (PyExc_ValueError, "Key not found");
		goto EXIT;
	}

	//	Create numpy array that point to C global data
	//	
	if (ptr->object == NULL) {

		PyObject * ndarray;

		// Structured array
		ndarray = PyArray_NewFromDescr (
			& PyArray_Type,
			*(ptr->type),
			ptr->ndim,
			ptr->dims,
			NULL,
			ptr->data,
			0,
			NULL
		);
		if (ndarray == NULL) goto EXIT;

		// create a view from an ndarray as a recarray of ptr->type
	
		ptr->object = PyArray_View ((PyArrayObject *) ndarray, NULL, recarray_type); 
	}

	Py_INCREF (ptr->object);

	return ptr->object;
	
EXIT:	return NULL;
}


////////////////////////////////////////////////////////////////////////////////
//
//	Return calibration catalog filename
//
static 
PyObject * calib_get_catalog (PyObject * self, PyObject * args)
{
	if (! PyArg_ParseTuple (args, ""))
		return NULL;

	return Py_BuildValue ("s", Get_catalog_filename());
}


////////////////////////////////////////////////////////////////////////////////
//
//	Test purpose function
//
static
PyObject * calib_test (PyObject * self, PyObject * args)
{
	PyObject *	obj;
	double		d;
	time_t		t;

	if (! PyArg_ParseTuple (args, "O", & obj)) {
		return NULL;
	}

	if (! PyDateTime_Check (obj)) {
		PyErr_SetString (PyExc_TypeError, "Argument must be datetime.datetime");
		return NULL;
	}

	d = PyFloat_AsDouble (PyObject_CallMethod (obj, "timestamp", ""));
	
	t = (time_t) d;

	if (Error (Read_calib ("AZS", t))) return NULL;
	
	return Py_None;
}
	

////////////////////////////////////////////////////////////////////////////////
//
//	List of library callable methods
//
static	PyMethodDef fonctions [] = {

	{ "init",		calib_init,		METH_VARARGS,	"init() -- Initialise calibration catalog\n"  },
	{ "read",		calib_get,		METH_VARARGS,	"read(datetime, key) -- return numpy array\n"},
	{ "catalog",		calib_get_catalog,	METH_VARARGS,	"catalog() -- return calibration catalog filename\n"},
	{ "test",		calib_test,		METH_VARARGS,	"test() -- test purpose function\n"},
	{ NULL, NULL, 0, NULL }
};


const char module_docstring [] = "Solar Orbiter Calibration library\n";


////////////////////////////////////////////////////////////////////////////////
//
//	Initial loading of the library
//
// 	Separate behaviour for Python 3.x and Python 2.x
//
#if PY_MAJOR_VERSION >= 3

	static struct PyModuleDef moduledef = {
		PyModuleDef_HEAD_INIT,
		"calib",
		module_docstring,
		-1,
		fonctions,
	};
			
	#define INITERROR return NULL

	PyObject *PyInit_calib(void)
	{
		PyObject *module = PyModule_Create (& moduledef);

		PyDateTime_IMPORT;

		import_array ();

		if (PyObject_CallMethod (module, "init", "") == NULL) return NULL;
			
		return module;
	}
#else
	#define	INITERROR return

	PyMODINIT_FUNC init_calib (void)
	{
		(void) Py_InitModule ("calib", fonctions);

		import_array ();
	}
#endif
