""""calib.py - SolarOrbiter PAS calibrations handling

Use PAS_calib low level C interface to read calibration files
"""
import numpy as np
import PAS_calib as _calib

import	datetime

UTC = datetime.timezone.utc

#	Define an AngleBin (amin, amoy, amax, sin, cos)
#
AngleBin = np.dtype ([
	("amin", "f4"),
	("amoy", "f4"),
	("amax", "f4"),
	("sin", "f4"),
	("cos", "f4")])

#	Define EnergyBin (emin, emoy, emax)
#
EnergyBin = np.dtype ([
	("emin", "f4"),
	("emoy", "f4"), 
	("emax", "f4")])


def	make_bins (data, type):
	"""
	Reshape array in array of angle or energy bins
	"""
	return np.rec.array (data, dtype = type).reshape (data.shape[:-1])


_func = {
	"AZS" : lambda dt, key :  make_bins (_calib.read (dt, "AZS"), AngleBin),
	"AZF" : lambda dt, key :  make_bins (_calib.read (dt, "AZF"), AngleBin),

	"ELS" : lambda dt, key :  make_bins (_calib.read (dt, "ELS"), AngleBin),
	"ELF" : lambda dt, key :  make_bins (_calib.read (dt, "ELF"), AngleBin),

	"EVS" : lambda dt, key :  make_bins (_calib.read (dt, "EVS"), EnergyBin),
	"EVF" : lambda dt, key :  make_bins (_calib.read (dt, "EVF"), EnergyBin),

	"GFF" : lambda dt, key :  _calib.read (dt, "GFF"),
	"CNF" : lambda dt, key :  _calib.read (dt, "CNF"),
	"YIL" : lambda dt, key :  _calib.read (dt, "YIL"),
	"ELD" : lambda dt, key :  _calib.read (dt, "ELD"),
}

def	read (dt, key):
	"""
	Return calibration table for a given datetime and key

	dt : datetime, assumed to be UTC timetag
	"""
	return _func[key] (dt.replace (tzinfo = UTC), key)
	
	
def	catalog():
	"""
	Return fullpath of calibration catalog
	"""
	return _calib.catalog()


def	test():

	import 	datetime

	now = datetime.datetime.now()

	azs = read (now, "AZS")
	azf = read (now, "AZF")

	els = read (now, "ELS")
	elf = read (now, "ELF")

	evs = read (now, "EVS")
	evf = read (now, "EVF")

	gff = read (now, "GFF")
	cnf = read (now, "CNF")
	yil = read (now, "YIL")
	eld = read (now, "ELD")

	print ("Simplified azimuth table (CEM)")

	for i in range (azs.shape[0]):
		print ("Azimuth[%2d] = % 7.3f +/- %5.3f" % (i, azs[i].amoy, (azs[i].amax - azs[i].amin)/2.0))

	print ("Simplified elevation table")

	for i, (amin, amoy, amax, sin, cos) in enumerate (els):
		print ("Elevation[%1d] = % 7.3f +/- %5.3f" % (i, amoy, (amax - amin)/2.0))

	print ("Simplified energy table")

	for i, (emin, emoy, emax) in enumerate (evs):
		print ("Energy[%2d] = %10.3f +/-[%7.3f, %7.3f]" % (i, emoy, emax - emoy, emoy - emin))

	print ("gff", gff.shape)
	print ("cnf", cnf.shape)
	print ("yil", yil.shape)
	print ("eld", eld.shape)

if __name__ == "__main__":

	test()
