////////////////////////////////////////////////////////////////////////////////
//
//	Module	CALIB
//
//	SWA PAS calibration parameters
//
////////////////////////////////////////////////////////////////////////////////

#define	MODULE_NAME	"CALIB"

#include <math.h>

#include "TOOLS.h"
#include "CATALOG.h"
#include "CALIB.h"

#define	DEG_TO_RAD(deg) (M_PI / 180.0 * (deg))


////////////////////////////////////////////////////////////////////////////////
//
//	Compare to FILL_VALUE
//
int	Fill_value (float v)
{
	return fabs (v / FILL_VALUE) > 0.999;
}


////////////////////////////////////////////////////////////////////////////////
//
//	PAS Housekeeping calibrations (model depending)
//
//	Instrument model :
//	0 : EM
// 	1 : EQM
//	2 : PFM or FS
//
static	int		pas_model = 2;	

typedef struct {
	char *	varname;	// CDF varname
	char *	key;		// SCOS 2000 MIB identifier
	float 	coeff [3][3];	// Set of (a,b,c) coefficients f(x) = a.x^2 + b.x + c for (EM, EQM, PFM)
}	t_param;

static	t_param 	table [] = {

	{ "V_MON_C", "NIA00500", { 	
		{ 0.0, 1.221, 0.0 }, 
		{ 0.0, 1.221, 0.0 }, 
		{ 0.0, 1.221, 0.0 }
		}
	},
	{ "V_MON_L", "NIA00501", {
		{ 0.0, 1.221, 0.0 },
		{ 0.0, 1.221, 0.0 },
		{ 0.0, 1.221, 0.0 }
		}
	},
	{ "I_MON_C", "NIA00502", {
		{ 0.0, 0.122, 0.0 },
		{ 0.0, 0.122, 0.0 },
		{ 0.0, 0.122, 0.0 }
		}
	},
	{ "I_MON_L", "NIA00503", {
		{ 0.0, 0.122, 0.0 },
		{ 0.0, 0.122, 0.0 },
		{ 0.0, 0.122, 0.0 }
		}
	},
	{ "T_MON_C", "NIA00504", {
		{ 0.0, 0.0374, -52.06 },
		{ 0.0, 0.0378, -52.5 },
		{ 0.0, 0.0378, -52.6 }
		}
	},
	{ "T_MON_L", "NIA00505", { 
		{ 0.0, 0.0374, -52.06 },
		{ 0.0, 0.0378, -52.5 },
		{ 0.0, 0.0378, -52.6 }
		}
	},
	{ "T1_HEATHER", "NIA00506", {
		{ 0.0, 0.0374, -52.06 },
		{ 0.0, 0.0378, -52.5 },
		{ 0.0, 0.0378, -52.6 }
		}
	},
	{ "T2_HEATHER", "NIA00507", {
		{ 0.0, 0.0374, -52.06 },
		{ 0.0, 0.0378, -52.5 },
		{ 0.0, 0.0378, -52.6 }
		}
	},
	{ "PLUS_24_V_CEM_OUT", "NIA00508", {
		{ 0.0, 0.007324, 0.0 },
		{ 0.0, 0.007327, 0.0 },
		{ 0.0, 0.007327, 0.0 }
		}
	},
	{ "PLUS_5_V_CEM_OUT", "NIA00509", {
		{ 0.0, 0.001512, 0.0 },
		{ 0.0, 0.001512, 0.0 },
		{ 0.0, 0.001514, 0.0 }
		}
	},
	{ "PLUS_12_V_HT_OUT", "NIA00510", {
		{ 0.0, 0.003667, 0.0 },
		{ 0.0, 0.003668, 0.0 },
		{ 0.0, 0.003668, 0.0 }
		}
	},
	{ "MINUS_12_V_HT_OUT", "NIA00511", {
		{ 0.0, 0.003678, 0.0 },
		{ 0.0, 0.003678, 0.0 },
		{ 0.0, 0.003678, 0.0 }
		}
	},
	{ "PLUS_4V_FPGA_OUT", "NIA00512", {
		{ 0.0, 0.001221, 0.0 },
		{ 0.0, 0.001221, 0.0 },
		{ 0.0, 0.001221, 0.0 }
		}
	},
	{ "2V5_FPGA_OUT", "NIA00513", {
		{ 0.0, 0.001221, 0.0 },
		{ 0.0, 0.001221, 0.0 },
		{ 0.0, 0.001221, 0.0 }
		}
	},
	{ "TEMP_DCDC", "NIA00514", {
		{ 0.0, 0.0374, -52.06 },
		{ 0.0, 0.0378, -52.4 },
		{ 0.0, 0.0378, -52.6 }
		}
	},
	{ "TEMP_FPGA", "NIA00515", {
		{ 0.0, 0.0374, -52.06 },
		{ 0.0, 0.0378, -52.4 },
		{ 0.0, 0.0378, -52.6 }
		}
	},
	{ "HK_I_PLUS_24V_CEM", "NIA00516", {
		{ 0.0, 0.04907, 0.0 },
		{ 0.0, 0.0492, -0.2541 },
		{ 0.0, 0.04909, 0.0 }
		}
	},
	{ "HK_I_PLUS_5V_CEM", "NIA00517", {
		{ 0.0, 0.0245, 0.0 },
		{ 0.0, 0.0276, 0.3416 },
		{ 0.0, 0.0276, 0.3416 }
		}
	},
	{ "HK_I_PLUS_12V_HT", "NIA00518", {
		{ 0.0, 0.0244, 0.0 },
		{ 0.0, 0.0244, -2.0058 },
		{ 0.0, 0.0244, -2.0058 }
		}
	},
	{ "HK_I_MINUS_12V_HT", "NIA00519", {
		{ 0.0, 0.0244, 0.0 },
		{ 0.0, 0.0245, 0.0701 },
		{ 0.0, 0.02442, 0.0 }
		}
	},
	{ "HK_I_PLUS_5V_FPGA", "NIA00520", {
		{ 0.0, 0.0217, 0.0 },
		{ 0.0, 0.025, 0.0 },
		{ 0.0, 0.02171, 0.0 }
		}
	},
	{ "HK_I_PLUS_28V_PRI", "NIA00521", {
		{ 0.0, 0.229, -181.4 },
		{ -8e-06, 0.107, 3.88 },
		{ -6e-06, 0.1474, 0.0 }
		}
	},
	{ "HK_I_2V5_FPGA", "NIA00522", {
		{ 0.0, 0.081, 0.0 },
		{ 0.0, 0.0793, 0.408 },
		{ 0.0, 0.0793, 0.408 }
		}
	},
	{ "T3_HEATHER", "NIA00523", {
		{ 0.0, 0.0374, -52.06 },
		{ 0.0, 0.0378, -52.4 },
		{ 0.0, 0.0378, -52.6 }
		}
	},
	{ "TEMP_HVPS", "NIA00524", {
		{ 0.0, 0.054, -80.4 },
		{ 0.0, 0.0455, -55.93 },
		{ 0.0, 0.0455, -55.93 }
		}
	},
	{ "TEMP_EA", "NIA00525", {
		{ 0.0, 0.0357, -54.3 },
		{ 0.0, 0.0546, -80.82 },
		{ 0.0, 0.0546, -80.82 }
		}
	},
	{ "HK_MHV_POS", "NIA00526", {
		{ 0.0, 1.762, 0.0 },
		{ 0.0, 1.587, 0.0 },
		{ 0.0, 1.587, 0.0 }
		}
	},
	{ "HK_MHV_NEG", "NIA00527", {
		{ 0.0, 1.762, 0.0 },
		{ 0.0, 1.587, 0.0 },
		{ 0.0, 1.587, 0.0 }
		}
	},
	{ "HK_ANL", "NIA00528", {
		{ 0.0, 0.34, 0.0 },
		{ 0.0, 0.366, 0.0 },
		{ 0.0, 0.391, 0.0 }
		}
	},
	{ "HK_TOP_DEFL", "NIA00529", {
		{ 0.0, 1.356, 0.0 },
		{ 0.0, 1.367, 0.0 },
		{ 0.0, 1.356, 0.0 }
		}
	},
	{ "HK_TOP_CAP", "NIA00530", {
		{ 0.0, 0.678, 0.0 },
		{ 0.0, 1.367, 0.0 },
		{ 0.0, 1.364, 0.0 }
		}
	},
	{ "HK_BOT_DEFL", "NIA00531", {
		{ 0.0, 1.356, 0.0 },
		{ 0.0, 1.367, 0.0 },
		{ 0.0, 1.36, 0.0 }
		}
	},
	{ NULL, NULL }
};


////////////////////////////////////////////////////////////////////////////////
//
//	Return entry for a given HSK parameter key
//
static
t_param * Search_hsk_parameter (char * key)
{
	int	i;

	for (i = 0; table [i].key != NULL; i++) {

		if (strcasecmp (key, table[i].varname) == 0) return table + i;
	}
	return NULL;
}


////////////////////////////////////////////////////////////////////////////////
//
//	Physical value of HSK parameters
//
float	Hsk_physical_value (char * key, int value)
{
	char *		fname = FNAME ("Hsk_physical_value");
	t_param *	param = Search_hsk_parameter (key);
	float *		coeff;

	assert (param != NULL);

	coeff = param->coeff [pas_model];

	Message_log (3, fname, "KEY = %s COEFF = (%g, %g,%g)", 
		key,
		coeff[0],
		coeff[1],
		coeff[2]);

	return (coeff [0] * value + coeff [1]) * value + coeff [2];
}


////////////////////////////////////////////////////////////////////////////////
//
//	New PAS calibrations
//	
#define	BUFSIZE	512

static	char	buffer [BUFSIZE];

t_angle_bin	AZS [NB_CEM];
t_angle_bin	AZF [NB_CEM][NB_ELEVATION];

float		CNF [NB_CEM][NB_ELEVATION][NB_ENERGY];

t_angle_bin	ELS [NB_ELEVATION];
t_angle_bin	ELF [NB_CEM][NB_ELEVATION];
float		ELD [NB_ENERGY];

t_energy_bin	EVS [NB_ENERGY];
t_energy_bin	EVF [NB_CEM][NB_ELEVATION][NB_ENERGY];

float 		GFF [NB_CEM][NB_ELEVATION];

float		YIL [NB_CEM][NB_ELEVATION][NB_CEM][NB_ELEVATION];

static FILE *	calib_file = NULL;			// calibration file descriptor

////////////////////////////////////////////////////////////////////////////////
//
//	Open calibration filename
//
static
t_err	Open_calib_file (char * filename)
{
	char *		fname = FNAME ("Open_calib_file");
	t_err		error = OK;

	if ((calib_file = fopen (filename, "r")) == NULL) {

		Error_log (fname, "Error opening %s", filename);
		error = ERR_open;
		goto EXIT;	
	}
EXIT:	return error;
}


////////////////////////////////////////////////////////////////////////////////
//
//	Close calib file descriptor
//
static	
t_err	Close_calib_file ()
{
	t_err	error = OK;

	if (calib_file != NULL) fclose (calib_file);

EXIT:	return error;
}


////////////////////////////////////////////////////////////////////////////////
//
//	Read next line from calibration descriptor
//
static
t_err	Read_next_line (char * buffer)
{
	char *		fname = FNAME ("Read_next_line");
	t_err		error = OK;

	while (Read_line (calib_file, BUFSIZE, buffer) == OK) {

		if (buffer [0] == '#') continue;
		if (strlen (buffer) == 0) continue;
		break;
	}

EXIT:	return error;
}


////////////////////////////////////////////////////////////////////////////////
//
//	Read angle bin from a string buffer
//
static
t_err	Read_angle_bin (int expected, char * buffer, t_angle_bin * ptr)
{
	char *		fname = FNAME ("Read_angle_bin");
	t_err		error = OK;
	int		count, num;
	float		max, min, _sin, _cos;
	char *		format = "%d %f %f %f %f";

	count = sscanf (buffer, format, & num, & min, & max, & _sin, & _cos);

	if (count != 5) {

		Error_log (fname, "%s", buffer);
		Error_log (fname, "%d/5 values extracted", count);
		error = ERR_read;
		goto EXIT;
	}
	
	if (num != expected) {

		Error_log (fname, "Read angle %d whereas %d was expected", num, expected);	
		error = ERROR;
		goto EXIT;
	}
	ptr->max = max;
	ptr->min = min;

	if (Fill_value (max) || Fill_value (min)) {

		ptr->moy = FILL_VALUE;
		ptr->sin = 0.0;
		ptr->cos = 0.0; 
	}
	else {
		ptr->moy = (max + min ) / 2.0;
		ptr->sin = sin (DEG_TO_RAD (ptr->moy));
		ptr->cos = cos (DEG_TO_RAD (ptr->moy));
	}
	
EXIT:	return error;
}


////////////////////////////////////////////////////////////////////////////////
//
//	Read energy bin from a string buffer
//
static
t_err	Read_energy_bin (int expected, char * buffer, t_energy_bin * ptr)
{
	char *		fname = FNAME ("Read_angle_bin");
	t_err		error = OK;
	int		count, num;
	float		max, min;
	char *		format = "%d %f %f";

	count = sscanf (buffer, format, & num, & min, & max);

	if (count != 3) {

		Error_log (fname, "%s", buffer);
		Error_log (fname, "%d/3 values extracted", count);
		error = ERR_read;
		goto EXIT;
	}
	
	if (num != expected) {

		Error_log (fname, "Read energy %d whereas %d was expected", num, expected);	
		error = ERROR;
		goto EXIT;
	}
	ptr->max = max;
	ptr->min = min;

	if (Fill_value (max) || Fill_value (min)) 
		ptr->moy = FILL_VALUE;
	else
		ptr->moy = sqrt (max * min);
	
EXIT:	return error;
}


////////////////////////////////////////////////////////////////////////////////
//
//	Read simplified Azimuth table
//
static
t_err	Read_AZS (char * filename)
{
	char *		fname = FNAME ("Read_AZS");
	t_err		error = OK;
	int		az;

	if (Error (error = Open_calib_file (filename))) goto EXIT;

	Message_log (1, fname, "Reading %s", filename);

	for (az = 0; az < NB_CEM; az++) {

		if (Error (error = Read_next_line (buffer))) goto EXIT;

		if (Error (error = Read_angle_bin (az, buffer, & AZS[az]))) goto EXIT;
	}

EXIT:	Close_calib_file();
	return error;	
}


////////////////////////////////////////////////////////////////////////////////
//
//	Read full Azimuth table
//
static
t_err	Read_AZF (char * filename)
{
	char *		fname = FNAME ("Read_AZF");
	t_err		error = OK;
	int		az, el, num;

	if (Error (error = Open_calib_file (filename))) goto EXIT;

	Message_log (1, fname, "Reading %s", filename);

	for (el = 0; el < NB_ELEVATION; el++) {

		if (Error (error = Read_next_line (buffer))) goto EXIT;

		if ((sscanf (buffer, "Elevation # %d", & num) != 1) || (num != el)) {

			Error_log (fname, "%s", buffer);
			Error_log (fname, "Error reading Elevation index #%d", el);
			error = ERROR;
			goto EXIT;
		}
		for (az = 0; az < NB_CEM; az++) {

			if (Error (error = Read_next_line (buffer))) goto EXIT;

			if (Error (error = Read_angle_bin (az, buffer, & AZF [az][el]))) goto EXIT;
		}
	}

EXIT:	Close_calib_file();
	return error;	
}


////////////////////////////////////////////////////////////////////////////////
//
//	Read calibration factor coefficients for 3D counts 
//
static
t_err	Read_CNF (char * filename)
{
	char *		fname = FNAME ("Read_CNF");
	t_err		error = OK;
	char *		format = "%d %f %f %f %f %f %f %f %f %f %f %f";
	int		el, az, en;
 
	if (Error (error = Open_calib_file (filename))) goto EXIT;

	Message_log (1, fname, "Reading %s", filename);

	for (en = 0; en < NB_ENERGY; en++)  {

		int	id, count;

		if (Error (error = Read_next_line (buffer))) goto EXIT;

		if ((count = sscanf (buffer, "%d", & id)) != 1) {
			Error_log (fname, "Could not find energy #0%2d : %s", en, buffer);
			error = ERR_read;
			goto EXIT;
		}

		if (id != en) {
			Error_log (fname, "Found energy %d, expecting %d", id, en);
			error = ERR_read;
			goto EXIT;
		}

		Message_log (3, fname, "Reading E%02d coefficients", en);

		for (el = 0; el < NB_ELEVATION; el++) {

			float	t [NB_CEM];

			if (Error (error = Read_next_line (buffer))) goto EXIT;

			count = sscanf (buffer, format, & id, t, t+1, t+2, t+3, t+4, t+5, t+6, t+7, t+8, t+9, t+10);

			Message_log (3, fname, "%d values for E%02d Elev %d", count, en, el);

			if (count != 12) {

				Error_log (fname, "Only %d values : %s", count, buffer);
				error = ERR_read;
				goto EXIT;
			}

			if (id != el) {
				Error_log (fname, "Found elevation %d for energy %d, expecting %d", id, en, el);
				error = ERR_read;
				goto EXIT;
			}

			for (az = 0; az < NB_CEM; az++) CNF [az][el][en] = t[az];
		}
	}

EXIT:	Close_calib_file();
	return error;	
}


////////////////////////////////////////////////////////////////////////////////
//
//	Read simplified elevation table
//
static
t_err	Read_ELS (char * filename)
{
	char *		fname = FNAME ("Read_ELS");
	t_err		error = OK;
	int		el;

	if (Error (error = Open_calib_file (filename))) goto EXIT;

	Message_log (1, fname, "Reading %s", filename);

	for (el = 0; el < NB_ELEVATION; el++) {

		if (Error (error = Read_next_line (buffer))) goto EXIT;

		if (Error (error = Read_angle_bin (el, buffer, & ELS[el]))) goto EXIT;
	}

EXIT:	Close_calib_file();
	return error;	
}


////////////////////////////////////////////////////////////////////////////////
//
//	Read full elevation table
//
static
t_err	Read_ELF (char * filename)
{
	char *		fname = FNAME ("Read_ELF");
	t_err		error = OK;
	int		az, el, num;

	if (Error (error = Open_calib_file (filename))) goto EXIT;

	Message_log (1, fname, "Reading %s", filename);

	for (az = 0; az < NB_CEM; az++) {

		if (Error (error = Read_next_line (buffer))) goto EXIT;

		if ((sscanf (buffer, "Azimuth # %d", & num) != 1) || (num != az)) {
			Error_log (fname, "%s", buffer);
			Error_log (fname, "Error reading Azimuth index #%d", az);
			error = ERROR;
			goto EXIT;
		}
		for (el = 0; el < NB_ELEVATION; el++) {

			if (Error (error = Read_next_line (buffer))) goto EXIT;

			if (Error (error = Read_angle_bin (el, buffer, & ELF[az][el]))) goto EXIT;
		}
	}

EXIT:	Close_calib_file();
	return error;	
}


////////////////////////////////////////////////////////////////////////////////
//
//	Read elevation deviation tabme
//
static
t_err	Read_ELD (char * filename)
{
	char *		fname = FNAME ("Read_ELD");
	t_err		error = OK;
	int		en;

	if (Error (error = Open_calib_file (filename))) goto EXIT;

	Message_log (1, fname, "Reading %s", filename);

	for (en = 0; en < NB_ENERGY; en++) {

		int	count, num, skip;

		if (Error (error = Read_next_line (buffer))) goto EXIT;

		count = sscanf (buffer, "%d %f %f", & num, & skip, ELD + en);

		if ((count != 3) || (num != en)) {
			Error_log (fname, "%s", buffer);
			Error_log (fname, "Error reading index #%d", en);
			error = ERROR;
			goto EXIT;
		}
	}

EXIT:	Close_calib_file ();
	return error;	
}


////////////////////////////////////////////////////////////////////////////////
//
//	Read simplified energy table
//
static
t_err	Read_EVS (char * filename)
{
	char *		fname = FNAME ("Read_EVS");
	t_err		error = OK;
	int		en;

	if (Error (error = Open_calib_file (filename))) goto EXIT;

	Message_log (1, fname, "Reading %s", filename);

	for (en = 0; en < NB_ENERGY; en++) {

		if (Error (error = Read_next_line (buffer))) goto EXIT;

		if (Error (error = Read_energy_bin (en, buffer, & EVS[en]))) goto EXIT;
	}

EXIT:	Close_calib_file();
	return error;	
}


////////////////////////////////////////////////////////////////////////////////
//
//	Read full energy table
//
static
t_err	Read_EVF (char * filename)
{
	char *		fname = FNAME ("Read_EVF");
	t_err		error = OK;
	int		en, az, el;
	char * 		format = "%d %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f";

	if (Error (error = Open_calib_file (filename))) goto EXIT;

	Message_log (1, fname, "Reading %s", filename);

	for (en = 0; en < NB_ENERGY; en++) {

		int	num, count;
		float	t [NB_CEM * 2];

		if (Error (error = Read_next_line (buffer))) goto EXIT;

		if ((sscanf (buffer, "%d", & num) != 1) || (num != en)) {

			Error_log (fname, "%s", buffer);
			Error_log (fname, "Error reading energy index #%d", en);
			error = ERROR;
			goto EXIT;
		}
	
		for (el = 0; el < NB_ELEVATION; el++) {

			if (Error (error = Read_next_line (buffer))) goto EXIT;

			count = sscanf (buffer, format, 
				& num, t, t+1, t+2, t+3, t+4, t+5, t+6, t+7, t+8, t+9, t+10, 
				t+11, t+12, t+13, t+14, t+15, t+16, t+17, t+18, t+19, t+20, t+21);

			if (count != 23) {
				Error_log (fname, "%s", buffer);
				Error_log (fname, "%d/23 values extracted", count);
				error = ERROR;
				goto EXIT;
			}

			if (num != el) {
				Error_log (fname, "%s", buffer);
				Error_log (fname, "Found elevation %d for energy %d whereas expecting %d", num, en, el);
				error = ERROR;
				goto EXIT;
			}

			for (az = 0; az < NB_CEM; az++) {

				float	min, max;
				min = t [az*2];
				max = t [az*2+1];

				EVF[az][el][en].min = min;
				EVF[az][el][en].max = max;

				EVF[az][el][en].moy = (Fill_value (min) || Fill_value (max))
					? FILL_VALUE
					: sqrt (min * max);
			}
		}
	}

EXIT:	Close_calib_file();
	return error;	
}


////////////////////////////////////////////////////////////////////////////////
//
//	Read geometric factor
//
static
t_err	Read_GFF (char * filename)
{
	char *		fname = FNAME ("Read_GFF");
	t_err		error = OK;
	char *		format = "%d %f %f %f %f %f %f %f %f %f %f %f";
	int		el;
 
	if (Error (error = Open_calib_file (filename))) goto EXIT;

	Message_log (1, fname, "Reading %s", filename);

	for (el = 0; el < NB_ELEVATION; el++) {

		int	az, num, count;
		float	t [NB_CEM];

		if (Error (error = Read_next_line (buffer))) goto EXIT;

		count = sscanf (buffer, format, & num, t, t+1, t+2, t+3, t+4, t+5, t+6, t+7, t+8, t+9, t+10);

		if (count != 12) {
			Error_log (fname, "%s", buffer);
			Error_log (fname, "Only %d values : %s", count, buffer);
			error = ERR_read;
			goto EXIT;
		}

		if (num != el) {
			Error_log (fname, "%s", buffer);
			Error_log (fname, "Found elevation %d whereas %d was expected", num, el);
			error = ERR_read;
			goto EXIT;
		}
		for (az = 0; az < NB_CEM; az++) GFF [az][el] = t[az];
	}

EXIT:	Close_calib_file();
	return error;	
}


////////////////////////////////////////////////////////////////////////////////
//
//	Read ghost correction table
//
static
t_err	Read_YIL (char * filename)
{
	char *		fname = FNAME ("Read_YIL");
	t_err		error = OK;
	int		az, el;
	int		num;

	if (Error (error = Open_calib_file (filename))) goto EXIT;

	Message_log (1, fname, "Reading %s", filename);

	for (az = 0; az < NB_CEM; az++) {

		if (Error (error = Read_next_line (buffer))) goto EXIT;

		if ((sscanf (buffer, "Azimuth_BIN # %d", & num) != 1) || (num != az)) {
			Error_log (fname, "%s", buffer);
			Error_log (fname, "Error reading Azimuth bin #%d", az);
			error = ERR_read;
			goto EXIT;
		}

		for (el = 0; el < NB_ELEVATION; el++) {

			int	cell;
			float	t [NB_ELEVATION];
			char *	format = "Azimuth_CELL # %d %f %f %f %f %f %f %f %f %f";

			if (Error (error = Read_next_line (buffer))) goto EXIT;

			if ((sscanf (buffer, "Elevation_BIN # %d", & num) != 1) || (num != el)) {
				Error_log (fname, "%s", buffer);
				Error_log (fname, "Error reading Elevation bin %d for azimuth %d", el, az);
				error = ERR_read;
				goto EXIT;
			}

			for (cell = 0; cell < NB_CEM; cell++) {

				float * t = & YIL [az][el][cell][0];
				int	count;

				if (Error (error = Read_next_line (buffer))) goto EXIT;
	
				count = sscanf (buffer, format, & num, t, t+1, t+2, t+3, t+4, t+5, t+6, t+7, t+8);

				if ((count != 10) || (num != cell)) {
					Error_log (fname, "%s", buffer);
					Error_log (fname, "Error Azimuth_cell %d for azmimuth %d, elevation %d", cell, az, el);
					error = ERR_read;
					goto EXIT;
				}
			}
		}
	}

EXIT:	Close_calib_file();
	return error;	
}


////////////////////////////////////////////////////////////////////////////////
//
//	Get calibration data for a given key and timetag
//
t_err	Read_calib (char * key, time_t time)
{
	char *		fname = FNAME ("Read_calib");
	t_err		error = OK;

	if (Get_calib_validity (key) < time) {

		if (Error (error = Read_catalog (key, time))) goto EXIT;
	}
	else	Message_log (3, fname, "%s calibration valid until %s", key, time_to_str (Get_calib_validity (key)));

EXIT:	return error;
}


////////////////////////////////////////////////////////////////////////////////
//
//	Set calibration parameters for a given model
//
t_err	Init_calib (char * model)
{
	char *		fname = FNAME ("Init_calib");
	t_err		error = OK;

	if	(strcasecmp (model, "EM") == 0) pas_model = 0;
	else 
	if	(strcasecmp (model, "EQM") == 0) pas_model = 1;
	else 
	if	(strcasecmp (model, "PFM") == 0) pas_model = 2;
	else 
	if	(strcasecmp (model, "FS") == 0) pas_model = 2;
	else {
		Error_log (fname, "Unknown model : %s", model);
		error = ERR_argument;
		goto EXIT;
	}

	Message_log (1, fname, "Model = %s => %d", model, pas_model);

	if (Error (error = Init_catalog())) goto EXIT;

	if (Error (error = Set_reader ("AZS", Read_AZS))) goto EXIT;
	if (Error (error = Set_reader ("AZF", Read_AZF))) goto EXIT;
	if (Error (error = Set_reader ("CNF", Read_CNF))) goto EXIT;
	if (Error (error = Set_reader ("ELS", Read_ELS))) goto EXIT;
	if (Error (error = Set_reader ("ELF", Read_ELF))) goto EXIT;
	if (Error (error = Set_reader ("ELD", Read_ELD))) goto EXIT;
	if (Error (error = Set_reader ("EVS", Read_EVS))) goto EXIT;
	if (Error (error = Set_reader ("EVF", Read_EVF))) goto EXIT;
	if (Error (error = Set_reader ("GFF", Read_GFF))) goto EXIT;
	if (Error (error = Set_reader ("YIL", Read_YIL))) goto EXIT;

EXIT:	return error;
}
